#ifdef __cplusplus
#include <NTL/ZZ.h>
#include <NTL/ZZX.h>
#include <NTL/ZZ_pXFactoring.h>
#include <NTL/ZZXFactoring.h>
#include <NTL/ZZ_p.h>
#include <NTL/ZZ_pX.h>
#include <NTL/ZZ_pE.h>
#include <NTL/ZZ_pEX.h>
#include <NTL/ZZ_pEXFactoring.h>
#include <NTL/lzz_p.h>
#include <NTL/lzz_pX.h>
#include <NTL/mat_ZZ.h>
#include <NTL/mat_poly_ZZ.h>
#include <NTL/GF2E.h>
#include <NTL/GF2X.h>
#include <NTL/GF2XFactoring.h>
#include <NTL/GF2EX.h>
#include <NTL/mat_GF2.h>
#include <NTL/mat_GF2E.h>
#include <NTL/HNF.h>
#include <NTL/LLL.h>
#include <gmp.h>
using namespace NTL;
#endif

#ifdef __cplusplus
#define EXTERN extern "C"
#else 
#define EXTERN
#endif

#include "Python.h"
#include "ccobject.h"

EXTERN void del_charstar(char*);

EXTERN void setup_NTL_error_callback(void (*function)(const char*, void*), void* context);

////////  ZZ //////////

#ifndef __cplusplus
typedef struct ZZ ZZ;
#endif

EXTERN int ZZ_to_int(const ZZ* x);
EXTERN ZZ* int_to_ZZ(int value);
EXTERN void ZZ_to_mpz(mpz_t* output, const ZZ* x);
EXTERN void mpz_to_ZZ(ZZ *output, const mpz_t* x);
EXTERN void ZZ_set_from_int(ZZ* x, int value);
/*Random-number generation */
//EXTERN void setSeed(const ZZ* n);
//EXTERN ZZ* ZZ_randomBnd(const ZZ* x);
//EXTERN ZZ* ZZ_randomBits(long n);

#ifdef __cplusplus
EXTERN long ZZ_remove(ZZ& x, const ZZ& a, const ZZ& p);
#endif

////////  ZZ_p //////////

#ifndef __cplusplus
typedef struct ZZ_p ZZ_p;
typedef struct ZZ_pContext ZZ_pContext;
#endif

#ifdef __cplusplus  // sorry, if you want a C version, feel free to add it
EXTERN int ZZ_p_to_int(const ZZ_p& x);
EXTERN ZZ_p int_to_ZZ_p(int value);
#endif
EXTERN void ZZ_p_set_from_int(ZZ_p* x, int value);
EXTERN ZZ_p* ZZ_p_pow(const ZZ_p* x, long e);
EXTERN void ntl_ZZ_set_modulus(ZZ* x);
EXTERN ZZ_p* ZZ_p_inv(ZZ_p* x);
EXTERN ZZ_p* ZZ_p_neg(ZZ_p* x);
EXTERN ZZ_p* ZZ_p_random(void);
EXTERN void ZZ_p_modulus(ZZ* mod, const ZZ_p* x);


EXTERN ZZ_pContext* ZZ_pContext_new(ZZ* p);
EXTERN ZZ_pContext* ZZ_pContext_construct(void* mem, ZZ* p);

//////// ZZX //////////
#ifndef __cplusplus
typedef struct ZZX ZZX;
#endif

EXTERN char* ZZX_repr(ZZX* x);
EXTERN ZZX* ZZX_copy(ZZX* x);
EXTERN void ZZX_setitem_from_int(ZZX* x, long i, int value);
EXTERN int ZZX_getitem_as_int(ZZX* x, long i);
EXTERN void ZZX_getitem_as_mpz(mpz_t* output, ZZX* x, long i);
EXTERN ZZX* ZZX_div(ZZX* x, ZZX* y, int* divisible);
EXTERN void ZZX_quo_rem(ZZX* x, ZZX* other, ZZX** r, ZZX** q);
EXTERN ZZX* ZZX_square(ZZX* x);
EXTERN int ZZX_equal(ZZX* x, ZZX* y);
EXTERN int ZZX_is_monic(ZZX* x);
EXTERN ZZX* ZZX_neg(ZZX* x);
EXTERN ZZX* ZZX_left_shift(ZZX* x, long n);
EXTERN ZZX* ZZX_right_shift(ZZX* x, long n);
EXTERN char* ZZX_content(ZZX* x);
EXTERN ZZX* ZZX_primitive_part(ZZX* x);
EXTERN void ZZX_pseudo_quo_rem(ZZX* x, ZZX* y, ZZX** r, ZZX** q);
EXTERN ZZX* ZZX_gcd(ZZX* x, ZZX* y);
EXTERN void ZZX_xgcd(ZZX* x, ZZX* y, ZZ** r, ZZX** s, ZZX** t, int proof);
EXTERN long ZZX_degree(ZZX* x);
EXTERN ZZ* ZZX_leading_coefficient(ZZX* x);
EXTERN char* ZZX_constant_term(ZZX* x);
EXTERN void ZZX_set_x(ZZX* x);
EXTERN int ZZX_is_x(ZZX* x);
EXTERN ZZX* ZZX_derivative(ZZX* x);
EXTERN ZZX* ZZX_reverse(ZZX* x);
EXTERN ZZX* ZZX_reverse_hi(ZZX* x, int hi);
EXTERN ZZX* ZZX_truncate(ZZX* x, long m);
EXTERN ZZX* ZZX_multiply_and_truncate(ZZX* x, ZZX* y, long m);
EXTERN ZZX* ZZX_square_and_truncate(ZZX* x, long m);
EXTERN ZZX* ZZX_invert_and_truncate(ZZX* x, long m);
EXTERN ZZX* ZZX_multiply_mod(ZZX* x, ZZX* y,  ZZX* modulus);
EXTERN ZZ* ZZX_trace_mod(ZZX* x, ZZX* y);
/* EXTERN ZZ* ZZX_polyeval(ZZX* f, ZZ* a); */
EXTERN char* ZZX_trace_list(ZZX* x);
EXTERN ZZ* ZZX_resultant(ZZX* x, ZZX* y, int proof);
EXTERN ZZ* ZZX_norm_mod(ZZX* x, ZZX* y, int proof);
EXTERN ZZ* ZZX_discriminant(ZZX* x, int proof);
EXTERN ZZX* ZZX_charpoly_mod(ZZX* x, ZZX* y, int proof);
EXTERN ZZX* ZZX_minpoly_mod(ZZX* x, ZZX* y);
EXTERN void ZZX_clear(ZZX* x);
EXTERN void ZZX_preallocate_space(ZZX* x, long n);

//////// ZZXFactoring //////////

// OUTPUT: v -- pointer to list of n ZZX elements (the squarefree factors)
//         e -- point to list of e longs (the exponents)
//         n -- length of above two lists
//  The lists v and e are mallocd, and must be freed by the calling code.
EXTERN void ZZX_squarefree_decomposition(ZZX*** v, long** e, long* n, ZZX* x);


//////// ZZ_pX //////////
#ifndef __cplusplus
typedef struct ZZ_pX ZZ_pX;
#endif

EXTERN ZZ_pX* ZZ_pX_init();
//EXTERN char* ZZ_pX_repr(ZZ_pX* x);
/* EXTERN ZZ_pX* ZZ_pX_copy(ZZ_pX* x); */
/* EXTERN void ZZ_pX_setitem_from_int(ZZ_pX* x, long i, int value); */
/* EXTERN int ZZ_pX_getitem_as_int(ZZ_pX* x, long i); */
/* EXTERN ZZ_pX* ZZ_pX_div(ZZ_pX* x, ZZ_pX* y, int* divisible); */
/* EXTERN ZZ_pX* ZZ_pX_mod(ZZ_pX* x, ZZ_pX* y); */
/* EXTERN void ZZ_pX_quo_rem(ZZ_pX* x, ZZ_pX* other, ZZ_pX** r, ZZ_pX** q); */
/* EXTERN ZZ_pX* ZZ_pX_square(ZZ_pX* x); */
/* EXTERN int ZZ_pX_equal(ZZ_pX* x, ZZ_pX* y); */
/* EXTERN int ZZ_pX_is_monic(ZZ_pX* x); */
/* EXTERN ZZ_pX* ZZ_pX_neg(ZZ_pX* x); */
/* EXTERN ZZ_pX* ZZ_pX_left_shift(ZZ_pX* x, long n); */
/* EXTERN ZZ_pX* ZZ_pX_right_shift(ZZ_pX* x, long n); */
/* EXTERN void ZZ_pX_quo_rem(ZZ_pX* x, ZZ_pX* y, ZZ_pX** r, ZZ_pX** q); */
/* EXTERN ZZ_pX* ZZ_pX_gcd(ZZ_pX* x, ZZ_pX* y); */
/* EXTERN void ZZ_pX_xgcd(ZZ_pX** d, ZZ_pX** s, ZZ_pX** t, ZZ_pX* a, ZZ_pX* b); */
/* EXTERN void ZZ_pX_plain_xgcd(ZZ_pX** d, ZZ_pX** s, ZZ_pX** t, ZZ_pX* a, ZZ_pX* b); */
/* EXTERN long ZZ_pX_degree(ZZ_pX* x); */
/* EXTERN void ZZ_pX_set_x(ZZ_pX* x); */
/* EXTERN int ZZ_pX_is_x(ZZ_pX* x); */
/* EXTERN ZZ_pX* ZZ_pX_derivative(ZZ_pX* x); */
/* EXTERN ZZ_pX* ZZ_pX_reverse(ZZ_pX* x); */
/* EXTERN ZZ_pX* ZZ_pX_reverse_hi(ZZ_pX* x, int hi); */
/* EXTERN ZZ_pX* ZZ_pX_truncate(ZZ_pX* x, long m); */
/* EXTERN ZZ_pX* ZZ_pX_multiply_and_truncate(ZZ_pX* x, ZZ_pX* y, long m); */
/* EXTERN ZZ_pX* ZZ_pX_square_and_truncate(ZZ_pX* x, long m); */
/* EXTERN ZZ_pX* ZZ_pX_invert_and_truncate(ZZ_pX* x, long m); */
/* EXTERN ZZ_pX* ZZ_pX_multiply_mod(ZZ_pX* x, ZZ_pX* y,  ZZ_pX* modulus); */
/* EXTERN ZZ_p* ZZ_pX_trace_mod(ZZ_pX* x, ZZ_pX* y); */
EXTERN char* ZZ_pX_trace_list(ZZ_pX* x);
/* EXTERN ZZ_p* ZZ_pX_resultant(ZZ_pX* x, ZZ_pX* y); */
/* EXTERN ZZ_p* ZZ_pX_norm_mod(ZZ_pX* x, ZZ_pX* y); */
/* EXTERN ZZ_pX* ZZ_pX_charpoly_mod(ZZ_pX* x, ZZ_pX* y); */
/* EXTERN ZZ_pX* ZZ_pX_minpoly_mod(ZZ_pX* x, ZZ_pX* y); */
/* EXTERN void ZZ_pX_clear(ZZ_pX* x); */
// EXTERN void ZZ_pX_preallocate_space(ZZ_pX* x, long n);

// Factoring elements of ZZ_pX:
// OUTPUT: v -- pointer to list of n ZZ_pX elements (the irred factors)
//         e -- point to list of e longs (the exponents)
//         n -- length of above two lists
//  The lists v and e are mallocd, and must be freed by the calling code.
EXTERN void ZZ_pX_factor(ZZ_pX*** v, long** e, long* n, ZZ_pX* x, long verbose);
EXTERN void ZZ_pX_linear_roots(ZZ_p*** v, long* n, ZZ_pX* f);

#ifdef __cplusplus
EXTERN void ZZ_pX_conv_modulus(ZZ_pX &fout, const ZZ_pX &fin, const ZZ_pContext &mod);
EXTERN void ZZ_pEX_conv_modulus(ZZ_pEX &fout, const ZZ_pEX &fin, const ZZ_pContext &mod);
EXTERN void ZZ_pX_min_val_coeff(long &valuation, long &index, const ZZ_pX &f, const ZZ &p);
EXTERN long ZZ_pX_get_val_coeff(const ZZ_pX &f, const ZZ &p, long i);
EXTERN void ZZ_pX_left_pshift(ZZ_pX &x, const ZZ_pX &a, const ZZ &pn, const ZZ_pContext &c);
EXTERN void ZZ_pX_right_pshift(ZZ_pX &x, const ZZ_pX &a, const ZZ &pn, const ZZ_pContext &c);
EXTERN void ZZ_pX_InvMod_newton_unram(ZZ_pX &x, const ZZ_pX &a, const ZZ_pXModulus &F, const ZZ_pContext &cpn, const ZZ_pContext &cp);
EXTERN void ZZ_pX_InvMod_newton_ram(ZZ_pX &x, const ZZ_pX &a, const ZZ_pXModulus &F, const ZZ_pContext &cpn);

#endif

//////// zz_p //////////

#ifndef __cplusplus
typedef struct zz_p zz_p;
typedef struct zz_pContext zz_pContext;
#endif

#define zz_p_set_from_long( obj1, obj2 )\
        (obj1) = (obj2)
#define NTL_zz_p_DOUBLE_EQUALS( obj1, obj2 )\
        (obj1) == (obj2)

EXTERN zz_pContext* zz_pContext_new(long p);
EXTERN zz_pContext* zz_pContext_construct(void* mem, long p);
EXTERN void zz_pContext_restore(zz_pContext* ctx);

//////// zz_pX //////////

#ifndef __cplusplus
typedef struct zz_pX zz_pX;
#endif 

#define NTL_zz_pX_DOUBLE_EQUALS( obj1, obj2 )\
        (obj1) == (obj2)

//////// ZZ_pEContext ///////////////

#ifndef __cplusplus
typedef struct ZZ_pEContext ZZ_pEContext;
#endif

EXTERN ZZ_pEContext* ZZ_pEContext_new(ZZ_pX *f);
EXTERN ZZ_pEContext* ZZ_pEContext_construct(void* mem, ZZ_pX *f);
EXTERN void ZZ_pEContext_restore(ZZ_pEContext* ctx);

//////// ZZ_pE ////////////

#ifndef __cplusplus
typedef struct ZZ_pE ZZ_pE;
#endif

EXTERN ZZ_pX ZZ_pE_to_ZZ_pX(ZZ_pE x);

//////// ZZ_pEX /////////

#ifndef __cplusplus
typedef struct ZZ_pEX ZZ_pEX;
#endif

//////// mat_ZZ //////////

#ifndef __cplusplus
typedef struct mat_ZZ mat_ZZ;
#endif

EXTERN void mat_ZZ_SetDims(mat_ZZ* mZZ, long nrows, long ncols);
EXTERN mat_ZZ* mat_ZZ_pow(const mat_ZZ* x, long e);
EXTERN long mat_ZZ_nrows(const mat_ZZ* x);
EXTERN long mat_ZZ_ncols(const mat_ZZ* x);
EXTERN void mat_ZZ_setitem(mat_ZZ* x, int i, int j, const ZZ* z);
EXTERN ZZ* mat_ZZ_getitem(const mat_ZZ* x, int i, int j);
EXTERN ZZ* mat_ZZ_determinant(const mat_ZZ* x, long deterministic);
EXTERN mat_ZZ* mat_ZZ_HNF(const mat_ZZ* A, const ZZ* D);
EXTERN ZZX* mat_ZZ_charpoly(const mat_ZZ* A);
EXTERN long mat_ZZ_LLL(ZZ **det, mat_ZZ *x, long a, long b, long verbose);
EXTERN long mat_ZZ_LLL_U(ZZ **det, mat_ZZ *x, mat_ZZ *U, long a, long b, long verbose);

/* //////// ZZ_p ////////// */
/* #ifndef __cplusplus */
/* typedef struct ZZ_p ZZ_p; */
/* #endif */

/* EXTERN void ZZ_p_set_modulus(const ZZ* p); */
/* EXTERN ZZ_p* new_ZZ_p(void); */
/* EXTERN void del_ZZ_p(ZZ_p* x); */
/* EXTERN ZZ_p* ZZ_p_add(const ZZ_p* x, const ZZ_p* y); */
/* EXTERN ZZ_p* ZZ_p_sub(const ZZ_p* x, const ZZ_p* y); */
/* EXTERN ZZ_p* ZZ_p_mul(const ZZ_p* x, const ZZ_p* y); */
/* EXTERN ZZ_p* ZZ_p_pow(const ZZ_p* x, long e); */
/* EXTERN int ZZ_p_is_zero(ZZ_p*x ); */
/* EXTERN int ZZ_p_is_one(ZZ_p*x ); */


//////// ZZ_pE //////////
#ifndef __cplusplus
typedef struct ZZ_pE ZZ_pE;
#endif

// EXTERN ZZ_pE* new_ZZ_pE



//////// ZZ_pEX //////////

//#ifndef __cplusplus
//typedef struct ZZ_pEX ZZ_pEX;
//#endif

//EXTERN ZZ_pEX* new_ZZ_pEX

/////// GF2X ////////////////
#ifndef __cplusplus
typedef struct GF2X GF2X;
#endif

/////// GF2EContext ////////////////

#ifndef __cplusplus
typedef struct GF2E GF2E;
typedef struct GF2EContext GF2EContext;
#endif

EXTERN GF2EContext* GF2EContext_new(GF2X* p);
EXTERN GF2EContext* GF2EContext_construct(void *mem, const GF2X *p);

//////// mat_GF2E //////////

#ifndef __cplusplus
typedef struct mat_GF2E mat_GF2E;
#endif

EXTERN void mat_GF2E_setitem(mat_GF2E* x, int i, int j, const GF2E* z);

//////// mat_GF2 //////////

#ifndef __cplusplus
typedef struct GF2 GF2;
typedef struct mat_GF2 mat_GF2;
#endif

EXTERN void mat_GF2_setitem(mat_GF2* x, int i, int j, const GF2* z);
